@@ -69,7 +69,7 @@ fn create_ui(app: &gtk::Application) {
     vbox.append(&label);
 
     window.set_child(Some(&vbox));
-    window.show();
+    window.present();
 
     app.add_window(&window);
 
