@@ -9,10 +9,10 @@
 //
 // SPDX-License-Identifier: MPL-2.0
 
-use gst_video::prelude::*;
-
-#[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
+use crate::sink::imp::PaintableSink;
+use crate::sink::imp::CAT;
 use gst_gl::prelude::*;
+use gst_video::subclass::prelude::*;
 use gtk::{gdk, glib};
 use std::{
     collections::{HashMap, HashSet},
@@ -64,7 +64,6 @@ impl VideoInfo {
 #[derive(Debug, PartialEq, Eq, Hash)]
 pub enum TextureCacheId {
     Memory(usize),
-    #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
     GL(usize),
     #[cfg(all(target_os = "linux", feature = "dmabuf"))]
     DmaBuf([i32; 4]),
@@ -76,7 +75,6 @@ enum MappedFrame {
         frame: gst_video::VideoFrame<gst_video::video_frame::Readable>,
         orientation: Orientation,
     },
-    #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
     GL {
         frame: gst_gl::GLVideoFrame<gst_gl::gl_video_frame::Readable>,
         wrapped_context: gst_gl::GLContext,
@@ -100,7 +98,6 @@ impl MappedFrame {
     fn buffer(&self) -> &gst::BufferRef {
         match self {
             MappedFrame::SysMem { frame, .. } => frame.buffer(),
-            #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
             MappedFrame::GL { frame, .. } => frame.buffer(),
             #[cfg(all(target_os = "linux", feature = "dmabuf"))]
             MappedFrame::DmaBuf { buffer, .. } => buffer,
@@ -110,7 +107,6 @@ impl MappedFrame {
     fn width(&self) -> u32 {
         match self {
             MappedFrame::SysMem { frame, .. } => frame.width(),
-            #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
             MappedFrame::GL { frame, .. } => frame.width(),
             #[cfg(all(target_os = "linux", feature = "dmabuf"))]
             MappedFrame::DmaBuf { info, .. } => info.width(),
@@ -120,7 +116,6 @@ impl MappedFrame {
     fn height(&self) -> u32 {
         match self {
             MappedFrame::SysMem { frame, .. } => frame.height(),
-            #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
             MappedFrame::GL { frame, .. } => frame.height(),
             #[cfg(all(target_os = "linux", feature = "dmabuf"))]
             MappedFrame::DmaBuf { info, .. } => info.height(),
@@ -130,7 +125,6 @@ impl MappedFrame {
     fn format_info(&self) -> gst_video::VideoFormatInfo {
         match self {
             MappedFrame::SysMem { frame, .. } => frame.format_info(),
-            #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
             MappedFrame::GL { frame, .. } => frame.format_info(),
             #[cfg(all(target_os = "linux", feature = "dmabuf"))]
             MappedFrame::DmaBuf { info, .. } => info.format_info(),
@@ -140,7 +134,6 @@ impl MappedFrame {
     fn orientation(&self) -> Orientation {
         match self {
             MappedFrame::SysMem { orientation, .. } => *orientation,
-            #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
             MappedFrame::GL { orientation, .. } => *orientation,
             #[cfg(all(target_os = "linux", feature = "dmabuf"))]
             MappedFrame::DmaBuf { orientation, .. } => *orientation,
@@ -225,12 +218,60 @@ pub(crate) struct Texture {
 struct FrameWrapper(gst_video::VideoFrame<gst_video::video_frame::Readable>);
 impl AsRef<[u8]> for FrameWrapper {
     fn as_ref(&self) -> &[u8] {
-        self.0.plane_data(0).unwrap()
+        let raw = self.0.as_ptr();
+        let data = unsafe {
+            std::slice::from_raw_parts((*raw).map[0].data as *const u8, (*raw).map[0].size)
+        };
+        data
     }
 }
 
 fn video_format_to_memory_format(f: gst_video::VideoFormat) -> gdk::MemoryFormat {
     match f {
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::Yuy2 => gdk::MemoryFormat::G8b8g8r8422,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::Yvyu => gdk::MemoryFormat::G8r8g8b8422,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::Uyvy => gdk::MemoryFormat::B8g8r8g8422,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::Vyuy => gdk::MemoryFormat::R8g8b8g8422,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::Nv12 => gdk::MemoryFormat::G8B8r8420,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::Nv21 => gdk::MemoryFormat::G8R8b8420,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::Nv16 => gdk::MemoryFormat::G8B8r8422,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::Nv61 => gdk::MemoryFormat::G8R8b8422,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::Nv24 => gdk::MemoryFormat::G8B8r8444,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::P01010le => gdk::MemoryFormat::G10x6B10x6r10x6420,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::P01010be => gdk::MemoryFormat::G10x6B10x6r10x6420,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::P012Le => gdk::MemoryFormat::G12x4B12x4r12x4420,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::P012Be => gdk::MemoryFormat::G12x4B12x4r12x4420,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::P016Le => gdk::MemoryFormat::G16B16r16420,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::P016Be => gdk::MemoryFormat::G16B16r16420,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::Yuv9 => gdk::MemoryFormat::G8B8R8410,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::Yvu9 => gdk::MemoryFormat::G8R8B8410,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::Y41b => gdk::MemoryFormat::G8B8R8411,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::I420 => gdk::MemoryFormat::G8B8R8420,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::Yv12 => gdk::MemoryFormat::G8R8B8420,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::Y42b => gdk::MemoryFormat::G8B8R8422,
+        #[cfg(feature = "gtk_v4_20")]
+        gst_video::VideoFormat::Y444 => gdk::MemoryFormat::G8B8R8444,
         #[cfg(feature = "gtk_v4_14")]
         gst_video::VideoFormat::Bgrx => gdk::MemoryFormat::B8g8r8x8,
         #[cfg(feature = "gtk_v4_14")]
@@ -249,6 +290,59 @@ fn video_format_to_memory_format(f: gst_video::VideoFormat) -> gdk::MemoryFormat
     }
 }
 
+#[cfg(feature = "gtk_v4_20")]
+fn videoinfo_to_color_state(info: &gst_video::VideoInfo) -> Option<gdk::ColorState> {
+    let colorimetry = info.colorimetry();
+    let color_params = gdk::CicpParams::new();
+
+    let range = match (colorimetry.range(), info.is_rgb()) {
+        (gst_video::VideoColorRange::Range0_255, ..) => gdk::CicpRange::Full,
+        (gst_video::VideoColorRange::Range16_235, ..) => gdk::CicpRange::Narrow,
+        (.., true) => gdk::CicpRange::Full,
+        (.., false) => gdk::CicpRange::Narrow,
+    };
+    color_params.set_range(range);
+
+    let matrix = match (colorimetry.matrix(), info.is_rgb()) {
+        (gst_video::VideoColorMatrix::Unknown, true) => gst_video::VideoColorMatrix::Rgb.to_iso(),
+        (gst_video::VideoColorMatrix::Unknown, false) => {
+            gst_video::VideoColorMatrix::Bt709.to_iso()
+        }
+        _ => colorimetry.matrix().to_iso(),
+    };
+    color_params.set_matrix_coefficients(matrix);
+
+    let tf = match (colorimetry.transfer(), info.is_rgb()) {
+        (gst_video::VideoTransferFunction::Unknown, true) => {
+            gst_video::VideoTransferFunction::Srgb.to_iso()
+        }
+        (gst_video::VideoTransferFunction::Unknown, false) => {
+            gst_video::VideoTransferFunction::Bt709.to_iso()
+        }
+        _ => colorimetry.transfer().to_iso(),
+    };
+    color_params.set_transfer_function(tf);
+
+    let primaries = match (colorimetry.primaries(), info.is_rgb()) {
+        (gst_video::VideoColorPrimaries::Unknown, true) => {
+            gst_video::VideoColorPrimaries::Bt709.to_iso()
+        }
+        (gst_video::VideoColorPrimaries::Unknown, false) => {
+            gst_video::VideoColorPrimaries::Bt709.to_iso()
+        }
+        _ => colorimetry.primaries().to_iso(),
+    };
+    color_params.set_color_primaries(primaries);
+
+    match color_params.build_color_state() {
+        Ok(color_state) => Some(color_state),
+        Err(error) => {
+            println!("Could not build color state: {}", error);
+            None
+        }
+    }
+}
+
 fn video_frame_to_memory_texture(
     frame: gst_video::VideoFrame<gst_video::video_frame::Readable>,
     cached_textures: &mut HashMap<TextureCacheId, gdk::Texture>,
@@ -267,16 +361,43 @@ fn video_frame_to_memory_texture(
     let format = video_format_to_memory_format(frame.format());
     let width = frame.width();
     let height = frame.height();
-    let rowstride = frame.plane_stride()[0] as usize;
 
-    let texture = gdk::MemoryTexture::new(
-        width as i32,
-        height as i32,
-        format,
-        &glib::Bytes::from_owned(FrameWrapper(frame)),
-        rowstride,
-    )
-    .upcast::<gdk::Texture>();
+    let texture = {
+        #[cfg(feature = "gtk_v4_20")]
+        {
+            let mut builder = gdk::MemoryTextureBuilder::new()
+                .set_width(width as i32)
+                .set_height(height as i32)
+                .set_format(format);
+
+            if let Some(color_state) = videoinfo_to_color_state(frame.info()) {
+                builder = builder.set_color_state(&color_state);
+            }
+
+            for plane in 0..(frame.n_planes() as usize) {
+                builder = builder
+                    .set_offset(plane as u32, frame.plane_offset()[plane] as usize)
+                    .set_stride_for_plane(plane as u32, frame.plane_stride()[plane] as usize);
+            }
+
+            builder = builder.set_bytes(Some(&glib::Bytes::from_owned(FrameWrapper(frame))));
+
+            builder.build()
+        }
+        #[cfg(not(feature = "gtk_v4_20"))]
+        {
+            let stride = frame.plane_stride()[0] as usize;
+
+            gdk::MemoryTexture::new(
+                width as i32,
+                height as i32,
+                format,
+                &glib::Bytes::from_owned(FrameWrapper(frame)),
+                stride,
+            )
+        }
+        .upcast::<gdk::Texture>()
+    };
 
     cached_textures.insert(TextureCacheId::Memory(ptr), texture.clone());
     used_textures.insert(TextureCacheId::Memory(ptr));
@@ -284,7 +405,6 @@ fn video_frame_to_memory_texture(
     (texture, pixel_aspect_ratio)
 }
 
-#[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
 fn video_frame_to_gl_texture(
     frame: gst_gl::GLVideoFrame<gst_gl::gl_video_frame::Readable>,
     cached_textures: &mut HashMap<TextureCacheId, gdk::Texture>,
@@ -354,17 +474,38 @@ fn video_frame_to_gl_texture(
             };
             let sync_point = (*sync_meta.as_ptr()).data;
 
-            gdk::GLTextureBuilder::new()
-                .set_context(Some(gdk_context))
-                .set_id(texture_id as u32)
-                .set_width(width as i32)
-                .set_height(height as i32)
-                .set_format(format)
-                .set_sync(Some(sync_point))
-                .build_with_release_func(move || {
-                    // Unmap and drop the GStreamer GL texture once GTK is done with it and not earlier
-                    drop(frame);
-                })
+            let builder = {
+                #[cfg(feature = "gtk_v4_20")]
+                {
+                    let mut mut_builder = gdk::GLTextureBuilder::new()
+                        .set_context(Some(gdk_context))
+                        .set_id(texture_id as u32)
+                        .set_width(width as i32)
+                        .set_height(height as i32)
+                        .set_format(format)
+                        .set_sync(Some(sync_point));
+
+                    if let Some(color_state) = videoinfo_to_color_state(frame.info()) {
+                        mut_builder = mut_builder.set_color_state(&color_state);
+                    }
+                    mut_builder
+                }
+                #[cfg(not(feature = "gtk_v4_20"))]
+                {
+                    gdk::GLTextureBuilder::new()
+                        .set_context(Some(gdk_context))
+                        .set_id(texture_id as u32)
+                        .set_width(width as i32)
+                        .set_height(height as i32)
+                        .set_format(format)
+                        .set_sync(Some(sync_point))
+                }
+            };
+
+            builder.build_with_release_func(move || {
+                // Unmap and drop the GStreamer GL texture once GTK is done with it and not earlier
+                drop(frame);
+            })
         }
         #[cfg(not(feature = "gtk_v4_12"))]
         {
@@ -411,17 +552,35 @@ fn video_frame_to_dmabuf_texture(
         return Ok((texture.clone(), pixel_aspect_ratio));
     }
 
-    let builder = gdk::DmabufTextureBuilder::new();
-    builder.set_display(&gdk::Display::default().unwrap());
-    builder.set_fourcc(info.fourcc());
-    builder.set_modifier(info.modifier());
-    builder.set_width(width);
-    builder.set_height(height);
-    builder.set_n_planes(n_planes);
+    let mut builder = gdk::DmabufTextureBuilder::new()
+        .set_display(&gdk::Display::default().unwrap())
+        .set_fourcc(info.fourcc())
+        .set_modifier(info.modifier())
+        .set_width(width)
+        .set_height(height)
+        .set_n_planes(n_planes);
+
+    #[cfg(feature = "gtk_v4_20")]
+    {
+        // Convert DRM fourcc to Gst VideoFormat if possible so we know whether the format is RGB or
+        // YCbCr. This is needed for cases where the colorimetry is not fully set. The fallback uses
+        // the DMA_DRM format which we assume is YCbCr.
+        let color_state = match info.to_video_info() {
+            Ok(v) => videoinfo_to_color_state(&v),
+            Err(_) => videoinfo_to_color_state(info),
+        };
+        if color_state.is_some() {
+            builder = builder.set_color_state(color_state.as_ref());
+        }
+    }
+
     for plane in 0..(n_planes as usize) {
-        builder.set_fd(plane as u32, fds[plane]);
-        builder.set_offset(plane as u32, offsets[plane] as u32);
-        builder.set_stride(plane as u32, strides[plane] as u32);
+        unsafe {
+            builder = builder.set_fd(plane as u32, fds[plane]);
+        }
+        builder = builder
+            .set_offset(plane as u32, offsets[plane] as u32)
+            .set_stride(plane as u32, strides[plane] as u32);
     }
 
     let texture = unsafe {
@@ -453,7 +612,6 @@ impl Frame {
             MappedFrame::SysMem { frame, .. } => {
                 video_frame_to_memory_texture(frame, cached_textures, &mut used_textures)
             }
-            #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
             MappedFrame::GL {
                 frame,
                 wrapped_context,
@@ -535,15 +693,11 @@ impl Frame {
 
 impl Frame {
     pub(crate) fn new(
+        dbg_obj: &PaintableSink,
         buffer: &gst::Buffer,
         info: &VideoInfo,
         orientation: Orientation,
-        #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))] wrapped_context: Option<
-            &gst_gl::GLContext,
-        >,
-        #[allow(unused_variables)]
-        #[cfg(not(any(target_os = "macos", target_os = "windows", feature = "gst-gl")))]
-        wrapped_context: Option<&()>,
+        wrapped_context: Option<&gst_gl::GLContext>,
     ) -> Result<Self, gst::FlowError> {
         // Empty buffers get filtered out in show_frame
         debug_assert!(buffer.n_memory() > 0);
@@ -551,122 +705,150 @@ impl Frame {
         #[allow(unused_mut)]
         let mut frame = None;
 
+        // Check we received a buffer with dmabuf memory and if so do some checks before
+        // passing it onwards
+        #[cfg(all(target_os = "linux", feature = "dmabuf"))]
+        if frame.is_none()
+            && buffer
+                .peek_memory(0)
+                .is_memory_type::<gst_allocators::DmaBufMemory>()
         {
-            // Check we received a buffer with dmabuf memory and if so do some checks before
-            // passing it onwards
-            #[cfg(all(target_os = "linux", feature = "dmabuf"))]
-            if frame.is_none()
-                && buffer
-                    .peek_memory(0)
-                    .is_memory_type::<gst_allocators::DmaBufMemory>()
+            if let Some((vmeta, info)) =
+                Option::zip(buffer.meta::<gst_video::VideoMeta>(), info.dma_drm())
             {
-                if let Some((vmeta, info)) =
-                    Option::zip(buffer.meta::<gst_video::VideoMeta>(), info.dma_drm())
-                {
-                    let mut fds = [-1i32; 4];
-                    let mut offsets = [0; 4];
-                    let mut strides = [0; 4];
-                    let n_planes = vmeta.n_planes() as usize;
+                let mut fds = [-1i32; 4];
+                let mut offsets = [0; 4];
+                let mut strides = [0; 4];
+                let n_planes = vmeta.n_planes() as usize;
 
-                    let vmeta_offsets = vmeta.offset();
-                    let vmeta_strides = vmeta.stride();
+                let vmeta_offsets = vmeta.offset();
+                let vmeta_strides = vmeta.stride();
 
-                    for plane in 0..n_planes {
-                        let Some((range, skip)) =
-                            buffer.find_memory(vmeta_offsets[plane]..(vmeta_offsets[plane] + 1))
-                        else {
-                            break;
-                        };
+                for plane in 0..n_planes {
+                    let Some((range, skip)) =
+                        buffer.find_memory(vmeta_offsets[plane]..(vmeta_offsets[plane] + 1))
+                    else {
+                        break;
+                    };
 
-                        let mem = buffer.peek_memory(range.start);
-                        let Some(mem) = mem.downcast_memory_ref::<gst_allocators::DmaBufMemory>()
-                        else {
-                            break;
-                        };
+                    let mem = buffer.peek_memory(range.start);
+                    let Some(mem) = mem.downcast_memory_ref::<gst_allocators::DmaBufMemory>()
+                    else {
+                        break;
+                    };
 
-                        let fd = mem.fd();
-                        fds[plane] = fd;
-                        offsets[plane] = mem.offset() + skip;
-                        strides[plane] = vmeta_strides[plane] as usize;
-                    }
+                    let fd = mem.fd();
+                    fds[plane] = fd;
+                    offsets[plane] = mem.offset() + skip;
+                    strides[plane] = vmeta_strides[plane] as usize;
+                }
 
-                    // All fds valid?
-                    if fds[0..n_planes].iter().all(|fd| *fd != -1) {
-                        frame = Some(MappedFrame::DmaBuf {
-                            buffer: buffer.clone(),
-                            info: info.clone(),
-                            n_planes: n_planes as u32,
-                            fds,
-                            offsets,
-                            strides,
-                            width: vmeta.width(),
-                            height: vmeta.height(),
-                            orientation,
-                        });
-                    }
+                // All fds valid?
+                if fds[0..n_planes].iter().all(|fd| *fd != -1) {
+                    frame = Some(MappedFrame::DmaBuf {
+                        buffer: buffer.clone(),
+                        info: info.clone(),
+                        n_planes: n_planes as u32,
+                        fds,
+                        offsets,
+                        strides,
+                        width: vmeta.width(),
+                        height: vmeta.height(),
+                        orientation,
+                    });
                 }
             }
-            #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
-            {
-                if frame.is_none() {
-                    // Check we received a buffer with GL memory and if the context of that memory
-                    // can share with the wrapped context around the GDK GL context.
-                    //
-                    // If not it has to be uploaded to the GPU.
-                    let memory_ctx = buffer
-                        .peek_memory(0)
-                        .downcast_memory_ref::<gst_gl::GLBaseMemory>()
-                        .and_then(|m| {
-                            let ctx = m.context();
-                            if wrapped_context
-                                .is_some_and(|wrapped_context| wrapped_context.can_share(ctx))
-                            {
-                                Some(ctx)
-                            } else {
-                                None
-                            }
-                        });
+        }
 
-                    if let Some(memory_ctx) = memory_ctx {
-                        // If there is no GLSyncMeta yet then we need to add one here now, which requires
-                        // obtaining a writable buffer.
-                        let mapped_frame = if buffer.meta::<gst_gl::GLSyncMeta>().is_some() {
-                            gst_gl::GLVideoFrame::from_buffer_readable(buffer.clone(), info)
-                                .map_err(|_| gst::FlowError::Error)?
-                        } else {
-                            let mut buffer = buffer.clone();
-                            {
-                                let buffer = buffer.make_mut();
-                                gst_gl::GLSyncMeta::add(buffer, memory_ctx);
-                            }
-                            gst_gl::GLVideoFrame::from_buffer_readable(buffer, info)
-                                .map_err(|_| gst::FlowError::Error)?
-                        };
-
-                        // Now that it's guaranteed that there is a sync meta and the frame is mapped, set
-                        // a sync point so we can ensure that the texture is ready later when making use of
-                        // it as gdk::GLTexture.
-                        let meta = mapped_frame.buffer().meta::<gst_gl::GLSyncMeta>().unwrap();
-                        meta.set_sync_point(memory_ctx);
-
-                        frame = Some(MappedFrame::GL {
-                            frame: mapped_frame,
-                            wrapped_context: wrapped_context.unwrap().clone(),
-                            orientation,
-                        });
+        if frame.is_none() {
+            // Check we received a buffer with GL memory and if the context of that memory
+            // can share with the wrapped context around the GDK GL context.
+            //
+            // If not it has to be uploaded to the GPU.
+            let memory_ctx = buffer
+                .peek_memory(0)
+                .downcast_memory_ref::<gst_gl::GLBaseMemory>()
+                .and_then(|m| {
+                    let ctx = m.context();
+                    if wrapped_context.is_some_and(|wrapped_context| wrapped_context.can_share(ctx))
+                    {
+                        Some(ctx)
+                    } else {
+                        None
                     }
-                }
+                });
+
+            if let Some(memory_ctx) = memory_ctx {
+                // If there is no GLSyncMeta yet then we need to add one here now, which requires
+                // obtaining a writable buffer.
+                let mapped_frame = if buffer.meta::<gst_gl::GLSyncMeta>().is_some() {
+                    gst_gl::GLVideoFrame::from_buffer_readable(buffer.clone(), info)
+                        .map_err(|_| gst::FlowError::Error)?
+                } else {
+                    let mut buffer = buffer.clone();
+                    {
+                        let buffer = buffer.make_mut();
+                        gst_gl::GLSyncMeta::add(buffer, memory_ctx);
+                    }
+                    gst_gl::GLVideoFrame::from_buffer_readable(buffer, info)
+                        .map_err(|_| gst::FlowError::Error)?
+                };
+
+                // Now that it's guaranteed that there is a sync meta and the frame is mapped, set
+                // a sync point so we can ensure that the texture is ready later when making use of
+                // it as gdk::GLTexture.
+                let meta = mapped_frame.buffer().meta::<gst_gl::GLSyncMeta>().unwrap();
+                meta.set_sync_point(memory_ctx);
+
+                frame = Some(MappedFrame::GL {
+                    frame: mapped_frame,
+                    wrapped_context: wrapped_context.unwrap().clone(),
+                    orientation,
+                });
             }
         }
 
         let mut frame = Self {
             frame: match frame {
                 Some(frame) => frame,
-                None => MappedFrame::SysMem {
-                    frame: gst_video::VideoFrame::from_buffer_readable(buffer.clone(), info)
-                        .map_err(|_| gst::FlowError::Error)?,
-                    orientation,
-                },
+                None => {
+                    let readable_frame = match buffer.n_memory() {
+                        1 => buffer.clone(),
+                        _ => {
+                            gst::debug!(CAT, imp = dbg_obj, "Buffer is not contiguous, copying");
+
+                            let mut copy_buffer = gst::Buffer::with_size(info.size()).unwrap();
+
+                            {
+                                let copy_buffer = copy_buffer.get_mut().unwrap();
+
+                                buffer
+                                    .copy_into(copy_buffer, gst::BUFFER_COPY_METADATA, 0..0)
+                                    .unwrap();
+
+                                let frame = gst_video::VideoFrameRef::from_buffer_ref_readable(
+                                    buffer, info,
+                                )
+                                .unwrap();
+                                let mut copy_frame =
+                                    gst_video::VideoFrameRef::from_buffer_ref_writable(
+                                        copy_buffer,
+                                        info,
+                                    )
+                                    .unwrap();
+
+                                frame.copy(&mut copy_frame).unwrap();
+                            }
+                            copy_buffer
+                        }
+                    };
+
+                    MappedFrame::SysMem {
+                        frame: gst_video::VideoFrame::from_buffer_readable(readable_frame, info)
+                            .map_err(|_| gst::FlowError::Error)?,
+                        orientation,
+                    }
+                }
             },
             overlays: vec![],
         };
