@@ -14,33 +14,18 @@ use crate::sink::frame::Frame;
 use crate::sink::paintable::Paintable;
 
 use glib::thread_guard::ThreadGuard;
-use gtk::glib;
 use gtk::prelude::*;
+use gtk::{gdk, glib};
 
-#[cfg(any(
-    target_os = "macos",
-    target_os = "windows",
-    feature = "gst-gl",
-    feature = "dmabuf"
-))]
-use gtk::gdk;
-
-#[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
-use gst_gl::prelude::GLContextExt as GstGLContextExt;
-#[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
-use gst_gl::prelude::*;
-
-#[allow(unused_imports)]
-use gst::prelude::*;
-use gst::subclass::prelude::*;
+use gst::{prelude::*, subclass::prelude::*};
 use gst_base::subclass::prelude::*;
-#[allow(unused_imports)]
-use gst_video::{prelude::*, subclass::prelude::*};
+use gst_gl::prelude::{GLContextExt as _, *};
+use gst_video::subclass::prelude::*;
 
-use once_cell::sync::Lazy;
+use std::sync::LazyLock;
 use std::sync::{
-    atomic::{AtomicBool, Ordering},
-    Mutex, MutexGuard,
+    atomic::{self, AtomicBool},
+    Mutex,
 };
 
 use crate::utils;
@@ -48,10 +33,10 @@ use crate::utils;
 // Global GL context that is created by the first sink and kept around until the end of the
 // process. This is provided to other elements in the pipeline to make sure they create GL contexts
 // that are sharing with the GTK GL context.
-#[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
 enum GLContext {
     Uninitialized,
     Unsupported,
+    #[allow(unused)]
     Initialized {
         display: gst_gl::GLDisplay,
         wrapped_context: gst_gl::GLContext,
@@ -59,10 +44,9 @@ enum GLContext {
     },
 }
 
-#[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
 static GL_CONTEXT: Mutex<GLContext> = Mutex::new(GLContext::Uninitialized);
 
-pub(crate) static CAT: Lazy<gst::DebugCategory> = Lazy::new(|| {
+pub(crate) static CAT: LazyLock<gst::DebugCategory> = LazyLock::new(|| {
     gst::DebugCategory::new(
         "gtk4paintablesink",
         gst::DebugColorFlags::empty(),
@@ -125,7 +109,7 @@ impl ObjectSubclass for PaintableSink {
 
 impl ObjectImpl for PaintableSink {
     fn properties() -> &'static [glib::ParamSpec] {
-        static PROPERTIES: Lazy<Vec<glib::ParamSpec>> = Lazy::new(|| {
+        static PROPERTIES: LazyLock<Vec<glib::ParamSpec>> = LazyLock::new(|| {
             vec![
                 glib::ParamSpecObject::builder::<super::paintable::Paintable>("paintable")
                     .nick("Paintable")
@@ -219,7 +203,7 @@ impl ObjectImpl for PaintableSink {
                 let mut settings = self.settings.lock().unwrap();
                 let value = value.get().expect("type checked upstream");
                 if settings.window_width != value {
-                    self.window_resized.store(true, Ordering::SeqCst);
+                    self.window_resized.store(true, atomic::Ordering::SeqCst);
                 }
                 settings.window_width = value;
             }
@@ -227,7 +211,7 @@ impl ObjectImpl for PaintableSink {
                 let mut settings = self.settings.lock().unwrap();
                 let value = value.get().expect("type checked upstream");
                 if settings.window_height != value {
-                    self.window_resized.store(true, Ordering::SeqCst);
+                    self.window_resized.store(true, atomic::Ordering::SeqCst);
                 }
                 settings.window_height = value;
             }
@@ -240,7 +224,7 @@ impl GstObjectImpl for PaintableSink {}
 
 impl ElementImpl for PaintableSink {
     fn metadata() -> Option<&'static gst::subclass::ElementMetadata> {
-        static ELEMENT_METADATA: Lazy<gst::subclass::ElementMetadata> = Lazy::new(|| {
+        static ELEMENT_METADATA: LazyLock<gst::subclass::ElementMetadata> = LazyLock::new(|| {
             gst::subclass::ElementMetadata::new(
                 "GTK 4 Paintable Sink",
                 "Sink/Video",
@@ -253,7 +237,7 @@ impl ElementImpl for PaintableSink {
     }
 
     fn pad_templates() -> &'static [gst::PadTemplate] {
-        static PAD_TEMPLATES: Lazy<Vec<gst::PadTemplate>> = Lazy::new(|| {
+        static PAD_TEMPLATES: LazyLock<Vec<gst::PadTemplate>> = LazyLock::new(|| {
             // Those are the supported formats by a gdk::Texture
             let mut caps = gst::Caps::new_empty();
             {
@@ -278,12 +262,24 @@ impl ElementImpl for PaintableSink {
                 }
 
                 for features in [
-                    #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
+                    #[cfg(any(
+                        feature = "x11egl",
+                        feature = "x11glx",
+                        feature = "waylandegl",
+                        target_os = "macos",
+                        target_os = "windows"
+                    ))]
                     Some(gst::CapsFeatures::new([
                         gst_gl::CAPS_FEATURE_MEMORY_GL_MEMORY,
                         gst_video::CAPS_FEATURE_META_GST_VIDEO_OVERLAY_COMPOSITION,
                     ])),
-                    #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
+                    #[cfg(any(
+                        feature = "x11egl",
+                        feature = "x11glx",
+                        feature = "waylandegl",
+                        target_os = "macos",
+                        target_os = "windows"
+                    ))]
                     Some(gst::CapsFeatures::new([
                         gst_gl::CAPS_FEATURE_MEMORY_GL_MEMORY,
                     ])),
@@ -296,11 +292,54 @@ impl ElementImpl for PaintableSink {
                     ])),
                     None,
                 ] {
-                    #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
                     {
                         const GL_FORMATS: &[gst_video::VideoFormat] =
                             &[gst_video::VideoFormat::Rgba, gst_video::VideoFormat::Rgb];
                         const NON_GL_FORMATS: &[gst_video::VideoFormat] = &[
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::Yuy2,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::Yvyu,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::Uyvy,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::Vyuy,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::Nv12,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::Nv21,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::Nv16,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::Nv61,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::Nv24,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::P01010le,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::P01010be,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::P012Le,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::P012Be,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::P016Le,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::P016Be,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::Yuv9,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::Yvu9,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::Y41b,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::I420,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::Yv12,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::Y42b,
+                            #[cfg(feature = "gtk_v4_20")]
+                            gst_video::VideoFormat::Y444,
                             #[cfg(feature = "gtk_v4_14")]
                             gst_video::VideoFormat::Bgrx,
                             #[cfg(feature = "gtk_v4_14")]
@@ -337,29 +376,6 @@ impl ElementImpl for PaintableSink {
                         }
                         caps.append(c);
                     }
-                    #[cfg(not(any(
-                        target_os = "macos",
-                        target_os = "windows",
-                        feature = "gst-gl"
-                    )))]
-                    {
-                        const FORMATS: &[gst_video::VideoFormat] = &[
-                            gst_video::VideoFormat::Bgra,
-                            gst_video::VideoFormat::Argb,
-                            gst_video::VideoFormat::Rgba,
-                            gst_video::VideoFormat::Abgr,
-                            gst_video::VideoFormat::Rgb,
-                            gst_video::VideoFormat::Bgr,
-                        ];
-
-                        let mut c = gst_video::video_make_raw_caps(FORMATS).build();
-
-                        if let Some(features) = features {
-                            let c = c.get_mut().unwrap();
-                            c.set_features_simple(Some(features));
-                        }
-                        caps.append(c);
-                    }
                 }
             }
 
@@ -424,7 +440,6 @@ impl ElementImpl for PaintableSink {
                 // Notify the pipeline about the GL display and wrapped context so that any other
                 // elements in the pipeline ideally use the same / create GL contexts that are
                 // sharing with this one.
-                #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
                 {
                     let gl_context = GL_CONTEXT.lock().unwrap();
                     if let GLContext::Initialized {
@@ -523,7 +538,7 @@ impl BaseSinkImpl for PaintableSink {
     }
 
     fn set_caps(&self, caps: &gst::Caps) -> Result<(), gst::LoggableError> {
-        gst::debug!(CAT, imp = self, "Setting caps {caps:?}");
+        gst::info!(CAT, imp = self, "Setting caps {caps:?}");
 
         #[allow(unused_mut)]
         let mut video_info = None;
@@ -567,7 +582,7 @@ impl BaseSinkImpl for PaintableSink {
                     settings.window_height
                 );
 
-                self.window_resized.store(false, Ordering::SeqCst);
+                self.window_resized.store(false, atomic::Ordering::SeqCst);
 
                 Some(
                     gst::Structure::builder("GstVideoOverlayCompositionMeta")
@@ -582,7 +597,6 @@ impl BaseSinkImpl for PaintableSink {
 
         query.add_allocation_meta::<gst_video::VideoOverlayCompositionMeta>(s.as_deref());
 
-        #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
         {
             if let GLContext::Initialized {
                 wrapped_context, ..
@@ -603,7 +617,6 @@ impl BaseSinkImpl for PaintableSink {
         gst::log!(CAT, imp = self, "Handling query {:?}", query);
 
         match query.view_mut() {
-            #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
             gst::QueryViewMut::Context(q) => {
                 // Avoid holding the locks while we respond to the query
                 // The objects are ref-counted anyway.
@@ -667,7 +680,7 @@ impl VideoSinkImpl for PaintableSink {
     fn show_frame(&self, buffer: &gst::Buffer) -> Result<gst::FlowSuccess, gst::FlowError> {
         gst::trace!(CAT, imp = self, "Rendering buffer {:?}", buffer);
 
-        if self.window_resized.swap(false, Ordering::SeqCst) {
+        if self.window_resized.swap(false, atomic::Ordering::SeqCst) {
             gst::debug!(CAT, imp = self, "Window size changed, needs to reconfigure");
             let obj = self.obj();
             let sink = obj.sink_pad();
@@ -694,11 +707,6 @@ impl VideoSinkImpl for PaintableSink {
             .unwrap_or(config.global_orientation);
 
         let wrapped_context = {
-            #[cfg(not(any(target_os = "macos", target_os = "windows", feature = "gst-gl")))]
-            {
-                None
-            }
-            #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
             {
                 let gl_context = GL_CONTEXT.lock().unwrap();
                 if let GLContext::Initialized {
@@ -711,10 +719,9 @@ impl VideoSinkImpl for PaintableSink {
                 }
             }
         };
-        let frame =
-            Frame::new(buffer, info, orientation, wrapped_context.as_ref()).map_err(|err| {
+        let frame = Frame::new(self, buffer, info, orientation, wrapped_context.as_ref())
+            .inspect_err(|_err| {
                 gst::error!(CAT, imp = self, "Failed to map video frame");
-                err
             })?;
         self.pending_frame.lock().unwrap().replace(frame);
 
@@ -810,7 +817,6 @@ impl PaintableSink {
             }
         }
 
-        #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
         {
             // Filter out GL caps from the template pads if we have no context
             if !matches!(&*GL_CONTEXT.lock().unwrap(), GLContext::Initialized { .. }) {
@@ -863,29 +869,40 @@ impl PaintableSink {
                 }
             });
 
-            #[cfg(feature = "gtk_v4_10")]
-            window.set_visible(true);
-            #[cfg(not(feature = "gtk_v4_10"))]
-            window.show();
+            window.present();
 
             *window_guard = Some(ThreadGuard::new(window));
         });
     }
 
-    fn create_paintable(&self, paintable_storage: &mut MutexGuard<Option<ThreadGuard<Paintable>>>) {
-        #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
+    fn create_paintable(&self, paintable_storage: &mut Option<ThreadGuard<Paintable>>) {
+        #[cfg(any(
+            feature = "x11egl",
+            feature = "x11glx",
+            feature = "waylandegl",
+            target_os = "macos",
+            target_os = "windows"
+        ))]
         {
             self.initialize_gl_context();
         }
+        #[cfg(not(any(
+            feature = "x11egl",
+            feature = "x11glx",
+            feature = "waylandegl",
+            target_os = "macos",
+            target_os = "windows"
+        )))]
+        {
+            gst::debug!(CAT, imp = self, "No GL platform enabled");
+            *GL_CONTEXT.lock().unwrap() = GLContext::Unsupported;
+        }
 
         self.configure_caps();
         self.initialize_paintable(paintable_storage);
     }
 
-    fn initialize_paintable(
-        &self,
-        paintable_storage: &mut MutexGuard<Option<ThreadGuard<Paintable>>>,
-    ) {
+    fn initialize_paintable(&self, paintable_storage: &mut Option<ThreadGuard<Paintable>>) {
         gst::debug!(CAT, imp = self, "Initializing paintable");
 
         // The channel for the SinkEvents
@@ -907,29 +924,27 @@ impl PaintableSink {
 
         // Create the paintable from the main thread
         let paintable = utils::invoke_on_main_thread(move || {
-            #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
-            {
-                let gdk_context = if let GLContext::Initialized { gdk_context, .. } =
-                    &*GL_CONTEXT.lock().unwrap()
-                {
+            let gdk_context =
+                if let GLContext::Initialized { gdk_context, .. } = &*GL_CONTEXT.lock().unwrap() {
                     Some(gdk_context.get_ref().clone())
                 } else {
                     None
                 };
-                ThreadGuard::new(Paintable::new(gdk_context))
-            }
-            #[cfg(not(any(target_os = "macos", target_os = "windows", feature = "gst-gl")))]
-            {
-                ThreadGuard::new(Paintable::new(None))
-            }
+            ThreadGuard::new(Paintable::new(gdk_context))
         });
 
-        **paintable_storage = Some(paintable);
+        *paintable_storage = Some(paintable);
 
         *self.sender.lock().unwrap() = Some(sender);
     }
 
-    #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
+    #[cfg(any(
+        feature = "x11egl",
+        feature = "x11glx",
+        feature = "waylandegl",
+        target_os = "macos",
+        target_os = "windows"
+    ))]
     fn initialize_gl_context(&self) {
         gst::debug!(CAT, imp = self, "Realizing GDK GL Context");
 
@@ -939,7 +954,13 @@ impl PaintableSink {
         });
     }
 
-    #[cfg(any(target_os = "macos", target_os = "windows", feature = "gst-gl"))]
+    #[cfg(any(
+        feature = "x11egl",
+        feature = "x11glx",
+        feature = "waylandegl",
+        target_os = "macos",
+        target_os = "windows"
+    ))]
     fn initialize_gl_context_main(&self) {
         gst::debug!(CAT, imp = self, "Realizing GDK GL Context from main thread");
 
@@ -978,25 +999,6 @@ impl PaintableSink {
             }
         };
 
-        match gdk_context.type_().name() {
-            #[cfg(feature = "x11egl")]
-            "GdkX11GLContextEGL" => (),
-            #[cfg(feature = "x11glx")]
-            "GdkX11GLContextGLX" => (),
-            #[cfg(feature = "waylandegl")]
-            "GdkWaylandGLContext" => (),
-            #[cfg(target_os = "macos")]
-            "GdkMacosGLContext" => (),
-            #[cfg(target_os = "windows")]
-            "GdkWin32GLContextWGL" => (),
-            #[cfg(all(windows, feature = "winegl"))]
-            "GdkWin32GLContextEGL" => (),
-            display => {
-                gst::error!(CAT, imp = self, "Unsupported GDK display {display} for GL");
-                return;
-            }
-        }
-
         gst::info!(CAT, imp = self, "Realizing GDK GL Context",);
 
         if let Err(err) = gdk_context.realize() {
@@ -1004,28 +1006,21 @@ impl PaintableSink {
             return;
         }
 
-        gst::info!(CAT, imp = self, "Successfully realized GDK GL Context");
+        gst::info!(
+            CAT,
+            imp = self,
+            "Successfully realized GDK GL Context of type {}",
+            gdk_context.type_().name()
+        );
 
         gdk_context.make_current();
 
-        let res = match gdk_context.type_().name() {
-            #[cfg(feature = "x11egl")]
-            "GdkX11GLContextEGL" => self.initialize_x11egl(gdk_display),
-            #[cfg(feature = "x11glx")]
-            "GdkX11GLContextGLX" => self.initialize_x11glx(gdk_display),
-            #[cfg(feature = "waylandegl")]
-            "GdkWaylandGLContext" => self.initialize_waylandegl(gdk_display),
-            #[cfg(target_os = "macos")]
-            "GdkMacosGLContext" => self.initialize_macosgl(gdk_display),
-            #[cfg(target_os = "windows")]
-            "GdkWin32GLContextWGL" => self.initialize_wgl(gdk_display, &gdk_context),
-            #[cfg(all(target_os = "windows", feature = "winegl"))]
-            "GdkWin32GLContextEGL" => self.initialize_winegl(gdk_display),
-            display_type => {
-                unreachable!("Unsupported GDK display {display_type} for GL");
-            }
+        let Some((display, wrapped_context)) = self.initialize_gst_gl(&gdk_display, &gdk_context)
+        else {
+            gst::warning!(CAT, imp = self, "Failed to initialize GStreamer GL context");
+            return;
         };
-        let (display, wrapped_context) = res.unwrap();
+
         match wrapped_context.activate(true) {
             Ok(_) => gst::info!(CAT, imp = self, "Successfully activated GL Context"),
             Err(_) => {
@@ -1060,10 +1055,47 @@ impl PaintableSink {
         };
     }
 
+    #[cfg(any(
+        feature = "x11egl",
+        feature = "x11glx",
+        feature = "waylandegl",
+        target_os = "macos",
+        target_os = "windows"
+    ))]
+    fn initialize_gst_gl(
+        &self,
+        gdk_display: &gdk::Display,
+        gdk_context: &gdk::GLContext,
+    ) -> Option<(gst_gl::GLDisplay, gst_gl::GLContext)> {
+        match gdk_context.type_().name() {
+            #[cfg(feature = "x11egl")]
+            "GdkX11GLContextEGL" => self.initialize_x11egl(gdk_display),
+            #[cfg(feature = "x11glx")]
+            "GdkX11GLContextGLX" => self.initialize_x11glx(gdk_display),
+            #[cfg(feature = "waylandegl")]
+            "GdkWaylandGLContext" => self.initialize_waylandegl(gdk_display),
+            #[cfg(target_os = "macos")]
+            "GdkMacosGLContext" => self.initialize_macosgl(gdk_display),
+            #[cfg(target_os = "windows")]
+            "GdkWin32GLContextWGL" => self.initialize_wgl(gdk_display, &gdk_context),
+            #[cfg(all(target_os = "windows", feature = "winegl"))]
+            "GdkWin32GLContextEGL" => self.initialize_winegl(gdk_display),
+            context_type => {
+                gst::error!(
+                    CAT,
+                    imp = self,
+                    "Unsupported GDK context {context_type} with display type {} for GL. This might be due to not having enabled a backend for GL when building the plugin",
+                    gdk_display.type_().name(),
+                );
+                None
+            }
+        }
+    }
+
     #[cfg(feature = "x11egl")]
     fn initialize_x11egl(
         &self,
-        display: gdk::Display,
+        display: &gdk::Display,
     ) -> Option<(gst_gl::GLDisplay, gst_gl::GLContext)> {
         gst::info!(
             CAT,
@@ -1084,7 +1116,7 @@ impl PaintableSink {
         unsafe {
             use glib::translate::*;
 
-            let display = display.downcast::<gdk_x11::X11Display>().unwrap();
+            let display = display.downcast_ref::<gdk_x11::X11Display>().unwrap();
             let x11_display =
                 gdk_x11::ffi::gdk_x11_display_get_egl_display(display.to_glib_none().0);
             if x11_display.is_null() {
@@ -1113,7 +1145,7 @@ impl PaintableSink {
     #[cfg(feature = "x11glx")]
     fn initialize_x11glx(
         &self,
-        display: gdk::Display,
+        display: &gdk::Display,
     ) -> Option<(gst_gl::GLDisplay, gst_gl::GLContext)> {
         gst::info!(
             CAT,
@@ -1134,7 +1166,7 @@ impl PaintableSink {
         unsafe {
             use glib::translate::*;
 
-            let display = display.downcast::<gdk_x11::X11Display>().unwrap();
+            let display = display.downcast_ref::<gdk_x11::X11Display>().unwrap();
             let x11_display = gdk_x11::ffi::gdk_x11_display_get_xdisplay(display.to_glib_none().0);
             if x11_display.is_null() {
                 gst::error!(CAT, imp = self, "Failed to get X11 display");
@@ -1162,7 +1194,7 @@ impl PaintableSink {
     #[cfg(feature = "waylandegl")]
     fn initialize_waylandegl(
         &self,
-        display: gdk::Display,
+        display: &gdk::Display,
     ) -> Option<(gst_gl::GLDisplay, gst_gl::GLContext)> {
         gst::info!(
             CAT,
@@ -1185,7 +1217,9 @@ impl PaintableSink {
 
             // let wayland_display = gdk_wayland::WaylandDisplay::wl_display(display.downcast());
             // get the ptr directly since we are going to use it raw
-            let display = display.downcast::<gdk_wayland::WaylandDisplay>().unwrap();
+            let display = display
+                .downcast_ref::<gdk_wayland::WaylandDisplay>()
+                .unwrap();
             let wayland_display =
                 gdk_wayland::ffi::gdk_wayland_display_get_wl_display(display.to_glib_none().0);
             if wayland_display.is_null() {
@@ -1216,7 +1250,7 @@ impl PaintableSink {
     #[cfg(target_os = "macos")]
     fn initialize_macosgl(
         &self,
-        display: gdk::Display,
+        display: &gdk::Display,
     ) -> Option<(gst_gl::GLDisplay, gst_gl::GLContext)> {
         gst::info!(
             CAT,
@@ -1253,7 +1287,7 @@ impl PaintableSink {
     #[cfg(target_os = "windows")]
     fn initialize_wgl(
         &self,
-        _display: gdk::Display,
+        _display: &gdk::Display,
         context: &gdk::GLContext,
     ) -> Option<(gst_gl::GLDisplay, gst_gl::GLContext)> {
         gst::info!(
@@ -1304,7 +1338,7 @@ impl PaintableSink {
     #[cfg(all(target_os = "windows", feature = "winegl"))]
     fn initialize_winegl(
         &self,
-        display: gdk::Display,
+        display: &gdk::Display,
     ) -> Option<(gst_gl::GLDisplay, gst_gl::GLContext)> {
         gst::info!(
             CAT,
@@ -1327,7 +1361,7 @@ impl PaintableSink {
             use gdk_win32::prelude::*;
             use glib::translate::*;
 
-            let d = display.downcast::<gdk_win32::Win32Display>().unwrap();
+            let d = display.downcast_ref::<gdk_win32::Win32Display>().unwrap();
             let egl_display = d.egl_display().unwrap().as_ptr();
 
             // TODO: On the binary distribution of GStreamer for Windows, this symbol is not there
